/* $Id: ast.h 10 2007-10-25 15:31:33Z csg $
  FILE:        ast.h
  CONTENTS:    ast.h   - This file contains the interface for the abstract syntax tree (AST).
  REVISION:    1.0, 960917, Christer Eriksson & Jukka Mki-Turja
  DESCRIPTION: Exports the constructors for the various types. The definition of the data
               type is present, but should be seen as private.
*/
#ifndef _AST_H_
#define _AST_H_


typedef struct treeNode *tTree;

/* Size of data types in trac42 */
# define INT_SIZE      1
# define BOOL_SIZE     1
# define STRING_SIZE 100

/* To distinguish between the node contents. */
typedef enum {
   kProgram,
   kFunction,
   kVariable,
   kAssign,
   kIf,
   kWhile,
   kRead,
   kWrite,
   kFuncCallStmnt,
   kReturn,
   kActual,
   kUnary,
   kBinary,
   kIntConst,
   kBoolConst,
   kStringConst,
   kFuncCallExpr,
   kRValue
} NODETYPE;

/* The NEG/SUB originates in the scanner where we not yet know which one it is. This unifiles them
   to the same value, which makes life a bit easier. */
#define MINUS 0

/* To distinguish between different kinds of unary expression nodes. */
typedef enum { NEG=MINUS, NOT } UNOP_KIND;

/* To distinguish between different kinds of binary expression nodes. */
typedef enum { SUB=MINUS, PLUS, MULT, DIV, OR, AND, EQ, LT, LE } BINOP_KIND;

/* To distinguish between formal parameters and local variables in variable nodes. */
typedef enum {kFormal, kLocal} vKind;

/* To distinguish between the data types in the trac42 language. */
typedef enum {VOID, BOOL, INT, STRING} eType;

/* The program. */
typedef struct {
   tTree Functions;  /* A list of functions in the program. */
} yProgram;

/* The function. */
typedef struct {
   tTree Next;
   tTree Variables;  /* Local variables as well as formal parameters. */
   tTree Stmnts;     /* A list of statements at the top level in the funktion. */
   char *Name;       /* The function name. */
   eType Type;       /* The return type. */
} yFunction;

/* The variable (local variable or formal parameter). */
typedef struct {
   tTree Next;
   vKind VarKind;    /* Tells if the variable is a local or a formal parameter. */
   char *Name;       /* The name of the variable. */
   eType Type;       /* The declared type. */
} yVariable;

/* A virtual Node to access the Next in every stmnt */
typedef struct {
   tTree Next;
} yStmnt;

/* The Assignment statement. */
typedef struct {
   tTree Next;
   char *Id;         /* The name of the variable on the left hand side of the assignment. */
   tTree Expr;       /* The expression on the right hand side. */
} yAssign;

/* The if statement. */
typedef struct {
   tTree Next;
   tTree Expr;       /* The conditional statement. */
   tTree Then;       /* The list of statements in the true branch of the if-statement. */
   tTree Else;       /* The list of statements in the false branch of the if-statement. */
} yIf;

/* The while statement. */
typedef struct {
   tTree Next;
   tTree Expr;       /* The conditional statement. */
   tTree Stmnt;      /* The list of statements in body of the while-statement. */
} yWhile;

/* The read statement. */
typedef struct {
   tTree Next;
   char *Id;         /* The name of the variable to be read to. */
} yRead;

/* The write statement. */
typedef struct {
   tTree Next;
   tTree Expr;       /* The expression whose value is to be written. */
} yWrite;

/* The return statement. */
typedef struct {
   tTree Next;
   tTree Expr;       /* The expression whose value is to be returned. */
} yReturn;

/* The function call statement. */
typedef struct {
   tTree Next;
   tTree Actuals;    /* A list of actual arguments passed to the called function. */
   char *FuncName;   /* The name of the called function. */
} yFuncCallStmnt;

/* The actual argument (this is just a wrapper to an expression). */
typedef struct {
   tTree Next;
   tTree Expr;       /* The expression of the actual argument. */
} yActual;

typedef struct {
} yExpr;

/* The unary expression. */
typedef struct {
   UNOP_KIND Operator; /* The operator type (! or -) */
   tTree Expr;             /* The sub-expression to operate on. */
} yUnary;

/* The binary expression. */
typedef struct {
   tTree LeftOperand;      /* The left hand side sub-expression to operate on. */
   BINOP_KIND Operator; /* The operator type */
   tTree RightOperand;     /* The right hand side sub-expression to operate on. */
} yBinary;

/* The int constant expression. */
typedef struct {
   int Value;        /* The integer value of the constant */
} yIntConst;

/* The bool constant expression. */
typedef struct {
   int Value;        /* The integer equivalent of the boolean value of the constant */
} yBoolConst;

/* The string constant expression. */
typedef struct {
   char *Value;     /* The string value of this constant. */
} yStringConst;

/* The function call expression. */
typedef struct {
   tTree Actuals;    /* A list of actual arguments to the function. */
   char *FuncName;  /* The name of the called function. */
} yFuncCallExpr;

/* The variable expression. */
typedef struct {
   char *Id;        /* The name of the variable. */
} yRValue;

struct treeNode {
   NODETYPE Kind;         /* Key to know which component in the union to select. */
   int LineNr;
   union {
      yProgram Program;
      yFunction Function;
      yVariable Variable;
      yStmnt Stmnt;
      yAssign Assign;
      yIf If;
      yWhile While;
      yRead Read;
      yWrite Write;
      yReturn Return;
      yFuncCallStmnt FuncCallStmnt;
      yExpr Expr;
      yFuncCallExpr FuncCallExpr;
      yActual Actual;
      yUnary Unary;
      yBinary Binary;
      yIntConst IntConst;
      yBoolConst BoolConst;
      yStringConst StringConst;
      yRValue RValue;
   } Node;
};

extern tTree treeRoot;

/* Returns a pointer to a program node. */
extern tTree mProgram(tTree pFunctions, int pLineNr);

/* Returns a pointer to a function node. */
extern tTree mFunction(tTree pVariables, tTree pStmnts, const char *pName, eType pType, int pLineNr);

/* Returns a pointer to a variable definition node. */
extern tTree mVariable(vKind pVarKind, const char *pName, eType pType, int pLineNr);

/* Returns a pointer to an assignment statement node. */
extern tTree mAssign(const char *pId, tTree pExpr, int pLineNr);

/* Returns a pointer to an if statement node. */
extern tTree mIf(tTree pExpr, tTree pThen, tTree pElse, int pLineNr);

/* Returns a pointer to a while statement node. */
extern tTree mWhile(tTree pExpr, tTree pStmnts, int pLineNr);

/* Returns a pointer to a read statement node. */
extern tTree mRead(const char *pId, int pLineNr);

/* Returns a pointer to a write statement node. */
extern tTree mWrite(tTree pExpr, int pLineNr);

/* Returns a pointer to a return statement node. */
extern tTree mReturn(tTree pExpr, int pLineNr);

/* Returns a pointer to a function call statement node. */
extern tTree mFuncCallStmnt(tTree pActuals, const char *pfuncName, int pLineNr);

/* Returns a pointer to a function call expression node. */
extern tTree mFuncCallExpr(tTree pActuals, const char *pfuncName, int pLineNr);

/* Returns a pointer to an actual argument node. */
extern tTree mActual(tTree pExpr, int pLineNr);

/* Returns a pointer to a unary expression node. */
extern tTree mUnary(UNOP_KIND pOperator, tTree pExpr, int pLineNr);

/* Returns a pointer to a binary expression node. */
extern tTree mBinary(tTree pLeftOperand, BINOP_KIND pOperator, tTree pRightOperand, int pLineNr);

/* Returns a pointer to an integer constant expression node. */
extern tTree mIntConst(int pValue, int pLineNr);

/* Returns a pointer to a bool constant expression node. */
extern tTree mBoolConst(int pValue, int pLineNr);

/* Returns a pointer to a string constant expression node. */
extern tTree mStringConst(const char *pValue, int pLineNr);

/* Returns a pointer to an rvalue(identifier) expression node. */
extern tTree mRValue(const char *pId, int pLineNr);

/* Returns a pointer to a list of functions where pFunction has been concatenated at the end of pFunctions. */
extern tTree connectFunctions(tTree pFunctions, tTree pFunction);

/* Returns a pointer to a list of variable nodes where second is a list of variables that has been concatenated at
   the end of the variable list first. The lists may be of any length. */
extern tTree connectVariables(tTree pVariables, tTree pVariable);

/* Returns a pointer to a list of statements where pStmnt has been concatenated at the end of pStmnts. */
extern tTree connectStmnts(tTree pStmnts, tTree pStmnt);

/* Returns a pointer to a list of actual arguments where pActual has been concatenated at the end of pActuals. */
extern tTree connectActuals(tTree pActuals, tTree pActual);

/* Returns a pointer to a list of variables consisting of the variables in pVariables updated with the type pType. */
extern tTree addType(tTree pVariables, eType pType);

/* To be used for cleanup when finished. */
extern void destroy_node(tTree node);

/* Returns a pointer to a copy of the string `s' in dynamic allocated memory. */
extern char *dstrcpy(const char *s);

#endif
