/* $Id: instruction.c 98 2008-01-08 18:12:03Z csg $ */
#include "instruction.h"
#include "vm.h"

#include "ast.h"
#include "emit.h"

#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <allegro.h>
#include <cgui/mem.h>

#define BUFSIZE 1000

struct t_inst {
   int (*Semantics)(struct t_inst *inst);
   int target_address;
   int size;
   int offset;
   int value;
   char *string;
   char *stack_label;
   char *src;
   t_vm *vm;
};

static char *WindToSpace(char *s)
{
   while (*s && !isspace(*s)) {
      s++;
   }
   return s;
}

static char *WindSpace(char *s)
{
   while (*s && isspace(*s)) {
      s++;
   }
   return s;
}

static char *WindToNextToken(char *s)
{
   s = WindToSpace(s);
   s = WindSpace(s);
   return s;
}

/* Clean up platform dependent new line charcters at the end */
static void FixupLine(char *buf)
{
   char *s;
   s = buf + strlen(buf)-1;
   while (s >= buf && isspace(*s)) {
      s--;
   }
   s[1] = 0;
}

static char *GetFirstToken(char *s)
{
   FixupLine(s);
   s = WindSpace(s);
   return s;
}

static void MatchHash(char **s)
{
   if (**s == '#') {
      (*s)++;
      *s = WindSpace(*s);
   }
}

/* These are the semanic functions for the instructions. However, they are just wrappers the
   actual semantics functions in the virtual machine. Their purpose is simply to unpack the
   oparands from our internal structure and pass them as parameter to the corresponding VM-function. */
static int InstrEnd(t_inst *inst)         { return VMEnd(inst->vm); }
static int InstrLabel(t_inst *inst)       { return VMLabel(inst->vm, inst); }
static int InstrLink(t_inst *inst)        { return VMLink(inst->vm); }
static int InstrUnlink(t_inst *inst)      { return VMUnlink(inst->vm); }
static int InstrRts(t_inst *inst)         { return VMRts(inst->vm); }
static int InstrDecl(t_inst *inst)        { return VMDecl(inst->vm, inst->size, inst->stack_label); }
static int InstrLval(t_inst *inst)        { return VMLval(inst->vm, &inst->offset, inst->stack_label); }
static int InstrLvalDynamic(t_inst *inst) { return VMLval(inst->vm, NULL, inst->stack_label); }
static int InstrAss(t_inst *inst)         { return VMAss(inst->vm, inst->size); }
static int InstrBrf(t_inst *inst)         { return VMBrf(inst->vm, inst->target_address); }
static int InstrBra(t_inst *inst)         { return VMBra(inst->vm, inst->target_address); }
static int InstrReadInt(t_inst *inst)     { return VMReadInt(inst->vm); }
static int InstrReadBool(t_inst *inst)    { return VMReadBool(inst->vm); }
static int InstrReadString(t_inst *inst)  { return VMReadString(inst->vm); }
static int InstrWriteInt(t_inst *inst)    { return VMWriteInt(inst->vm); }
static int InstrWriteBool(t_inst *inst)   { return VMWriteBool(inst->vm); }
static int InstrWriteString(t_inst *inst) { return VMWriteString(inst->vm); }
static int InstrBsr(t_inst *inst)         { return VMBsr(inst->vm, inst->target_address); }
static int InstrPop(t_inst *inst)         { return VMPop(inst->vm, inst->size); }
static int InstrNot(t_inst *inst)         { return VMNot(inst->vm); }
static int InstrNeg(t_inst *inst)         { return VMNeg(inst->vm); }
static int InstrOr(t_inst *inst)          { return VMOr(inst->vm); }
static int InstrAnd(t_inst *inst)         { return VMAnd(inst->vm); }
static int InstrEqString(t_inst *inst)    { return VMEqString(inst->vm); }
static int InstrEqInt(t_inst *inst)       { return VMEqInt(inst->vm); }
static int InstrLtString(t_inst *inst)    { return VMLtString(inst->vm); }
static int InstrLtInt(t_inst *inst)       { return VMLtInt(inst->vm); }
static int InstrLeString(t_inst *inst)    { return VMLeString(inst->vm); }
static int InstrLeInt(t_inst *inst)       { return VMLeInt(inst->vm); }
static int InstrSub(t_inst *inst)         { return VMSub(inst->vm); }
static int InstrAdd(t_inst *inst)         { return VMAdd(inst->vm); }
static int InstrMult(t_inst *inst)        { return VMMult(inst->vm); }
static int InstrDiv(t_inst *inst)         { return VMDiv(inst->vm); }
static int InstrPushInt(t_inst *inst)     { return VMPushInt(inst->vm, inst->value); }
static int InstrPushString(t_inst *inst)  { return VMPushString(inst->vm, inst->string); }
static int InstrRval(t_inst *inst)        { return VMRval(inst->vm, &inst->offset, inst->size); }
static int InstrRvalDynamic(t_inst *inst) { return VMRval(inst->vm, NULL, inst->size); }
static int InstrDeref(t_inst *inst)       { return VMDeref(inst->vm, inst->size, inst->stack_label); }
static int InstrAlloc(t_inst *inst)       { return VMAlloc(inst->vm, inst->size); }
static int InstrSetOffset(t_inst *inst)   { return VMSetOffset(inst->vm); }

static int LoadOperandFP(char **s, int *offset, int (**Semantics)(struct t_inst *inst))
{
   if (**s == '(') {
      if (*Semantics == InstrLval) {
         *Semantics = InstrLvalDynamic;
      } else {
         *Semantics = InstrRvalDynamic;
      }
   } else {
      MatchHash(s);
      sscanf(*s, "%d", offset);
   }
   return T42_LOAD_OK;
}

static int LoadOperandInteger(char **s, int *val)
{
   MatchHash(s);
   sscanf(*s, "%d", val);
   return T42_LOAD_OK;
}

static int LoadOperandString(char **s, char *strval)
{
   MatchHash(s);
   strcpy(strval, *s);
   *s += strlen(*s);
   return T42_LOAD_OK;
}

static int LoadOperandBool(char **s, int *val)
{
   char buf[100];

   MatchHash(s);
   *buf = 0;
   sscanf(*s, "%s", buf);
   if (strcmp(buf, "true") == 0) {
      *val = 1;
   } else if (strcmp(buf, "false") == 0) {
      *val = 0;
   } else {
      return T42_LOAD_FORMAT_ERROR;
   }
   return T42_LOAD_OK;
}

static int LoadComment(char **s, char *stack_label)
{
   int retcode=T42_LOAD_OK;
   *s = WindToNextToken(*s);
   if (**s) {
      if ((*s)[0] == '/' && (*s)[1] == '/') {
         /* Comment will be used as stack_label when possible and if present. */
         **s = 0;
         (*s) += 2;
         *s = WindSpace(*s);
         strcpy(stack_label, *s);
      } else {
         **s = 0;
         retcode = T42_LOAD_REDUNDANT_CHARS;
      }
   }
   return retcode;
}

static t_inst *InstructionCreate(t_vm *vm, int instr_address, int (*Semantics)(struct t_inst *inst), int offset, int size, int target_address, int intval, const char *strval, const char *stack_label, const char *src)
{
   t_inst *inst;

   inst = GetMem0(t_inst, 1);
   inst->vm = vm;
   inst->Semantics = Semantics;
   inst->offset = offset;
   inst->size = size;
   inst->target_address = target_address;
   inst->value = intval;
   if (strval) {
      inst->string = MkString(strval);
   }
   if (stack_label) {
      inst->stack_label = MkString(stack_label);
   }
   inst->src = MkString(src);
   VMAddInstruction(vm, instr_address, inst);
   return inst;
}

/* Scan the file using standard functions. */
extern int InstructionParseListLine(char *buf, t_vm *vm)
{
   char *s, instr_name[BUFSIZE]="", strval[BUFSIZE]="", stack_label[BUFSIZE]="";
   int instr_address=0, offset=0, size=0, target_address=0, intval=0, retcode=T42_LOAD_OK;
   int (*Semantics)(t_inst *inst);
   t_inst *inst;

   s = GetFirstToken(buf);
   if (*s) {
      /* Get target_address. */
      sscanf(s, "%d", &instr_address);
      s = WindToNextToken(s);
      /* Special handling of the lable - it fall out of the pattern. */
      if (*s == '[') {
         s++;
         sscanf(s, "%s", strval);
         strval[strlen(strval)-1] = 0;
         Semantics = InstrLabel;
         if (strcmp(strval, "trac42")==0) {
            VMSetTrac42Address(vm, instr_address);
         }
      } else {
         sscanf(s, "%s", instr_name);
         s = WindToNextToken(s);
         if (stricmp(instr_name, "LINK") == 0) {
            Semantics = InstrLink;
         } else if (stricmp(instr_name, "UNLINK") == 0) {
            Semantics = InstrUnlink;
         } else if (stricmp(instr_name, "RTS") == 0) {
            Semantics = InstrRts;
         } else if (stricmp(instr_name, "DECL") == 0) {
            Semantics = InstrDecl;
            retcode = LoadOperandInteger(&s, &size);
         } else if (stricmp(instr_name, "LVAL") == 0) {
            Semantics = InstrLval;
            retcode = LoadOperandFP(&s, &offset, &Semantics);
         } else if (stricmp(instr_name, "ASS") == 0) {
            Semantics = InstrAss;
            retcode = LoadOperandInteger(&s, &size);
         } else if (stricmp(instr_name, "ASSINT") == 0) {
            Semantics = InstrAss;
            size = 1;
         } else if (stricmp(instr_name, "ASSBOOL") == 0) {
            Semantics = InstrAss;
            size = 1;
         } else if (stricmp(instr_name, "ASSSTRING") == 0) {
            Semantics = InstrAss;
            size = STRING_SIZE;
         } else if (stricmp(instr_name, "BRF") == 0) {
            Semantics = InstrBrf;
            retcode = LoadOperandInteger(&s, &target_address);
         } else if (stricmp(instr_name, "BRA") == 0) {
            Semantics = InstrBra;
            retcode = LoadOperandInteger(&s, &target_address);
         } else if (stricmp(instr_name, "READINT") == 0) {
            Semantics = InstrReadInt;
         } else if (stricmp(instr_name, "READBOOL") == 0) {
            Semantics = InstrReadBool;
         } else if (stricmp(instr_name, "READSTRING") == 0) {
            Semantics = InstrReadString;
         } else if (stricmp(instr_name, "WRITEINT") == 0) {
            Semantics = InstrWriteInt;
         } else if (stricmp(instr_name, "WRITEBOOL") == 0) {
            Semantics = InstrWriteBool;
         } else if (stricmp(instr_name, "WRITESTRING") == 0) {
            Semantics = InstrWriteString;
         } else if (stricmp(instr_name, "BSR") == 0) {
            Semantics = InstrBsr;
            retcode = LoadOperandInteger(&s, &target_address);
         } else if (stricmp(instr_name, "POP") == 0) {
            Semantics = InstrPop;
            retcode = LoadOperandInteger(&s, &size);
         } else if (stricmp(instr_name, "NOT") == 0) {
            Semantics = InstrNot;
         } else if (stricmp(instr_name, "NEG") == 0) {
            Semantics = InstrNeg;
         } else if (stricmp(instr_name, "OR") == 0) {
            Semantics = InstrOr;
         } else if (stricmp(instr_name, "AND") == 0) {
            Semantics = InstrAnd;
         } else if (stricmp(instr_name, "EQINT") == 0) {
            Semantics = InstrEqInt;
         } else if (stricmp(instr_name, "EQBOOL") == 0) {
            Semantics = InstrEqInt;
         } else if (stricmp(instr_name, "EQSTRING") == 0) {
            Semantics = InstrEqString;
         } else if (stricmp(instr_name, "LTINT") == 0) {
            Semantics = InstrLtInt;
         } else if (stricmp(instr_name, "LTBOOL") == 0) {
            Semantics = InstrLtInt;
         } else if (stricmp(instr_name, "LTSTRING") == 0) {
            Semantics = InstrLtString;
         } else if (stricmp(instr_name, "LEINT") == 0) {
            Semantics = InstrLeInt;
         } else if (stricmp(instr_name, "LEBOOL") == 0) {
            Semantics = InstrLeInt;
         } else if (stricmp(instr_name, "LESTRING") == 0) {
            Semantics = InstrLeString;
         } else if (stricmp(instr_name, "SUB") == 0) {
            Semantics = InstrSub;
         } else if (stricmp(instr_name, "ADD") == 0) {
            Semantics = InstrAdd;
         } else if (stricmp(instr_name, "MULT") == 0) {
            Semantics = InstrMult;
         } else if (stricmp(instr_name, "DIV") == 0) {
            Semantics = InstrDiv;
         } else if (stricmp(instr_name, "PUSHINT") == 0) {
            Semantics = InstrPushInt;
            retcode = LoadOperandInteger(&s, &intval);
         } else if (stricmp(instr_name, "PUSHBOOL") == 0) {
            Semantics = InstrPushInt;
            retcode = LoadOperandBool(&s, &intval);
         } else if (stricmp(instr_name, "PUSHSTRING") == 0) {
            Semantics = InstrPushString;
            retcode = LoadOperandString(&s, strval);
            strval[STRING_SIZE-1] = 0;
         } else if (stricmp(instr_name, "RVAL") == 0) {
            Semantics = InstrRval;
            LoadOperandInteger(&s, &size);
            retcode = LoadOperandFP(&s, &offset, &Semantics);
         } else if (stricmp(instr_name, "RVALINT") == 0) {
            Semantics = InstrRval;
            size = 1;
            retcode = LoadOperandFP(&s, &offset, &Semantics);
         } else if (stricmp(instr_name, "RVALBOOL") == 0) {
            Semantics = InstrRval;
            size = 1;
            retcode = LoadOperandFP(&s, &offset, &Semantics);
         } else if (stricmp(instr_name, "RVALSTRING") == 0) {
            Semantics = InstrRval;
            size = STRING_SIZE;
            retcode = LoadOperandFP(&s, &offset, &Semantics);
         } else if (stricmp(instr_name, "DEREF") == 0) {
            Semantics = InstrDeref;
            retcode = LoadOperandInteger(&s, &size);
         } else if (stricmp(instr_name, "ALLOC") == 0) {
            Semantics = InstrAlloc;
            retcode = LoadOperandInteger(&s, &size);
         } else if (stricmp(instr_name, "SETOFFSET") == 0) {
            Semantics = InstrSetOffset;
         } else if (stricmp(instr_name, "END") == 0) {
            Semantics = InstrEnd;
         } else {
            return T42_LOAD_FORMAT_ERROR;
         }
         retcode = LoadComment(&s, stack_label);
      }
      if (retcode == T42_LOAD_OK) {
         inst = InstructionCreate(vm, instr_address, Semantics, offset, size, target_address, intval, strval, stack_label, buf);
      }
   }
   return retcode;
}

extern void InstructionDestroy(t_inst *instruction)
{
   free(instruction->string);
   free(instruction->stack_label);
   free(instruction->src);
   free(instruction);
}

extern void InstructionGetSource(t_inst *instruction, char **src)
{
   if (instruction->src) {
      *src = instruction->src;
   } else {
      *src = "";
   }
}

extern int InstructionExecute(t_inst *instruction)
{
   return instruction->Semantics(instruction);
}

extern t_inst *InstructionNop(t_vm *vm, int instr_address)
{
   char buf[100];
   sprintf(buf, "%3d   NOP", instr_address);
   return InstructionCreate(vm, instr_address, InstrLabel, 0, 0, 0, 0, NULL, NULL, buf);
}

extern t_inst *InstructionBsr(t_vm *vm, int instr_address, int target_address)
{
   char buf[100];
   sprintf(buf, "%3d   BSR", instr_address);
   return InstructionCreate(vm, instr_address, InstrBsr, 0, 0, target_address, 0, NULL, NULL, buf);
}

extern t_inst *InstructionEnd(t_vm *vm, int instr_address)
{
   char buf[100];
   sprintf(buf, "%3d   END", instr_address);
   return InstructionCreate(vm, instr_address, InstrEnd, 0, 0, 0, 0, NULL, NULL, buf);
}

extern char *InstructionGetStackLabel(t_inst *instruction)
{
   return instruction->stack_label;
}
