/* $Id: trac42i.c 98 2008-01-08 18:12:03Z csg $ */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <allegro.h>
#include <cgui.h>
#include <cgui/mem.h>
#include "ast.h"
#include "vm.h"

typedef struct t_stackline
{
   char regs[10];
   int address;
   int value;
   char *label;
} t_stackline;

typedef struct t_line {
   struct t_view *view;
   char *text;
   int isbreakpoint;
   int ispc;
   int i;
} t_line;

typedef struct t_view {
   t_line *line;
   int n;
   t_stackline *stacklines;
   int nstacklines;
   int idl;
   int idin;
   int idout;
   int idiow;
   int idw;
   int idstack;
   char *output;
   char input[STRING_SIZE];
   char *s;
   int input_in_progress;
   int exiting;
   int resetting;
   struct t_vm *vm;
} t_view;

static void DestroyCodeView(t_view *view)
{
   int i;

   for (i=0; i<view->n; i++) {
      Release(view->line[i].text);
   }
   Release(view->output);
   Release(view->stacklines);
   Release(view->line);
   Release(view);
}

static t_view *CreateCodeView(void)
{
   t_view *view;
   view = GetMem0(t_view, 1);
   return view;
}

#define MAXTAB 6
static void InitCodeView(t_view *view, struct t_vm *vm)
{
   int i, nchar, hnd;
   char *s, *p;

   view->vm = vm;
   view->n = VMNrOfInstructions(vm);
   view->line = GetMem0(t_line, view->n);
   view->output = MkString("");
   for (hnd=VMFirstInstruction(vm, &s), i=0; hnd!=END_INSTRUCTIONS; hnd=VMNextInstruction(hnd, vm, &s), i++) {
      p = view->line[i].text = GetMem(char, strlen(s)+MAXTAB+1);
      strcpy(p, s);
      if (*p && isspace(p[strlen(p)-1])) {
         p[strlen(p)-1] = 0;
      }
      for (; *p; p++) {
         if (*p=='\t') {
            *p = ' ';
            nchar = MAXTAB - (p - view->line[i].text);
            if (nchar > 0) {
               memmove(p+nchar, p, strlen(p)+1);
               memset(p, ' ', nchar);
            }
            break;
         }
      }
      view->line[i].view = view;
      view->line[i].i = i;
   }
}

void error_message(const char *message)
{
   set_gfx_mode(GFX_TEXT, 0, 0, 0, 0);
   allegro_message(message);
}

static void InputCallBack(void *data)
{
   t_view *view = data;
   int scan, ascii, offset;

   GetEditData(&scan, &ascii, &offset);
   if (ascii == '\r') {
      view->input_in_progress = 0;
      strcpy(view->s, view->input);
   }
}

static void CreateIOWindow(t_view *view)
{
   *view->output = 0;
   view->idiow = MkDialogue(ADAPTIVE, "Console window", W_FLOATING);
   view->idout = AddTextBox(DOWNLEFT, view->output, 200, 20, TB_FOCUS_END|TB_LINEFEED_|TB_WHITE|TB_FRAMESINK);
   view->idin = AddEditBox(DOWNLEFT, 200, "Input:", FSTRING, STRING_SIZE, view->input);
   DeActivate(view->idin);
   AddHandler(view->idin, InputCallBack, view);
   DisplayWin();
}

static void DisplayOutput(char *s, t_view *view)
{
   if (view->idout == 0) {
      CreateIOWindow(view);
   }
   view->output = ResizeMem(char, view->output, strlen(view->output) + strlen(s) + 2);
   if (*view->output) {
      strcat(view->output, "_");
   }
   strcat(view->output, s);
   UpdateTextBoxText(view->idout, view->output);
   Refresh(view->idout);
}

static void InputChecker(void *data)
{
   t_view *view = data;

   if (view->input_in_progress) {
      GenEvent(InputChecker, view, 0, view->idw);
   } else {
      DisplayOutput(view->input, view);
      *view->input = 0;
      DeActivate(view->idin);
      Refresh(view->idin);
      SetFocusOn(view->idw);
      SetFocusOn(view->idl);
      StopProcessEvents();
   }
}

static void GetInput(char *s, t_view *view)
{
   if (view->idin == 0) {
      CreateIOWindow(view);
   }
   view->input_in_progress = 1;
   view->s = s;
   InputChecker(view);
   Activate(view->idin);
   SetFocusOn(view->idin);
   ProcessEvents();
}

static void BuildStack(t_view *view)
{
   int rf, hnd;
   t_stackline *sl;

   Release(view->stacklines);
   view->stacklines = NULL;
   view->nstacklines = VMNrOfStackElements(view->vm);
   if (view->nstacklines < 0) {
      view->nstacklines = 0;
      return;
   }
   view->stacklines = sl = GetMem0(t_stackline, view->nstacklines);
   for (hnd=VMGetStackBottomElement(view->vm, &rf, &sl->address, &sl->value, &sl->label);
        hnd!=END_STACK_ELEMENTS;
       hnd=VMGetStackElementAbove(hnd, view->vm, &rf, &sl->address, &sl->value, &sl->label)) {
       *sl->regs = 0;
       if (rf & REG_FP) {
          strcat(sl->regs, "FP");
       }
       if (rf & REG_SP) {
          if (*sl->regs) {
             strcat(sl->regs, ",");
          }
          strcat(sl->regs, "SP");
       }
       sl++;
   }
}

static void ExitTrac42i(void *data)
{
   t_view *view = data;

   if (view->input_in_progress) {
      view->input_in_progress = 0;
      view->exiting = 1;
      simulate_keypress((KEY_F2<<8)|0);
      simulate_keypress(KEY_Q);
      return;
   }
   VMDestroy(view->vm);
   DestroyCodeView(view);
   StopProcessEvents();
}

static void CheckBounds(int status)
{
   switch (status) {
   case EXEC_ERROR_INSTR_CODE:
      Req("", "Unknown instruction recogniced.| OK ");
      break;
   case EXEC_ERROR_PC:
      Req("", "The vm counter is out of bounds.| OK ");
      break;
   case EXEC_ERROR_SP:
      Req("", "Stack overflow or underflow.| OK ");
      break;
   case EXEC_ERROR_DIV_ZERO:
      Req("", "Division by 0| OK ");
      break;
   case EXEC_CODE_END_SP_ERROR:
      Req("", "The stackpointer has not been restored to its initial state.| OK ");
      break;
   case EXEC_ERROR:
      Req("", "Unknown error.| OK ");
      break;
   case EXEC_CODE_END:
   case EXEC_OK:
   case EXEC_BREAK_POINT:
   default:
      break;
   }
}

static void StepCommand(void *data)
{
   t_view *view = data;
   int status;

   status = VMExecuteNextInstruction(view->vm);
   BuildStack(view);
   BrowseToL(view->idstack, view->nstacklines-1);
   BrowseTo(view->idl, VMGetPC(view->vm), 1);
   CheckBounds(status);
}

static void ExecuteCode(void *data)
{
   t_view *view = data;
   int status;

   if (view->exiting || view->resetting) {
      return;
   }
   status = VMExecuteNextInstruction(view->vm);
   if (status == EXEC_OK) {
      GenEvent(ExecuteCode, view, 0, view->idw);
   } else {
      BuildStack(view);
      Refresh(view->idstack);
      BrowseTo(view->idl, VMGetPC(view->vm), 0);
      CheckBounds(status);
   }
}

static void RunCommand(void *data)
{
   t_view *view = data;
   view->resetting = 0;
   ExecuteCode(view);
}

static void ResetCommand(void *data)
{
   t_view *view = data;

   view->resetting = 1;
   if (view->input_in_progress) {
      view->input_in_progress = 0;
      simulate_keypress((KEY_F2<<8)|0);
      return;
   }
   VMReset(view->vm);
   Remove(view->idiow);
   view->idiow = view->idin = view->idout = 0;
   BuildStack(view);
   Refresh(view->idstack);
   BrowseTo(view->idl, VMGetPC(view->vm), 0);
}

static void SetBreakPointCommand(void *data, int i)
{
   t_view *view = data;

   VMSetBreakPoint(i, view->vm);
   view->line[i].isbreakpoint = 1;
   Refresh(view->idl);
}

static void SetBreakPointCommandWrapper(void *data)
{
   t_view *view = data;
   SetFocusOn(view->idl);
   simulate_keypress((KEY_INSERT<<8)|0);
}

static void ClearBreakPointCommand(int id, void *data)
{
   t_line *line = data;
   (void)id;
   VMClearBreakPoint(line->i, line->view->vm);
   line->isbreakpoint = 0;
   Refresh(line->view->idl);
}

static void ClearAllBreakPointsCommand(void *data)
{
   t_view *view = data;
   int i;

   VMClearAllBreakPoints(view->vm);
   for (i=0; i<view->n; i++) {
      view->line[i].isbreakpoint = 0;
   }
   Refresh(view->idl);
}

static void ClearBreakPointCommandWrapper(void *data)
{
   t_view *view = data;
   SetFocusOn(view->idl);
   simulate_keypress((KEY_DEL<<8)|0);
}

static void CommandsMenu(void *data)
{
   t_view *view = data;

   MkMenuItem(0, "Step", "F7", StepCommand, view);
   MkMenuItem(0, "Run", "F9", RunCommand, view);
   MkMenuItem(0, "Reset", "F2", ResetCommand, view);
   MkMenuItem(0, "Set breakpoint", "Insert", SetBreakPointCommandWrapper, view);
   MkMenuItem(0, "Clear breakpoint", "Delete", ClearBreakPointCommandWrapper, view);
   MkMenuItem(0, "Clear all breakpoints", "", ClearAllBreakPointsCommand, view);
}

static void LoadFile(const char *fn, t_view *view)
{
   int retcode;
   char *ext;

   ext = strrchr(fn, '.');
   if (ext && strcmp(ext, ".obj")==0) {
      retcode = VMLoadBinFile(fn, view->vm);
   } else {
      retcode = VMLoadListFile(fn, view->vm);
   }
   switch (retcode) {
   case T42_LOAD_OPEN_FAIL:
      Request("", 0, 0, "We are not able to open the file %s. Anyway, we are really happy that you wanted to try!| OK ", fn);
      break;
   case T42_LOAD_NO_HANDLE:
      Req("", "The start handle \"trac42\" could not be found in the file. The start address has not been set. This means "
               "that you cannot run the program.| OK ");
      break;
   case T42_LOAD_FORMAT_ERROR:
      Request("", 0, 0, "There were one or more lines with unkown instructions in the list file %s, so we decided do "
                        "drop your code.| OK ", fn);
      InitCodeView(view, view->vm);
      Refresh(view->idl);
      break;
   case T42_LOAD_REDUNDANT_CHARS:
      Request("", 0, 0, "There were garbage at the end of some line(s) in %s| OK ", fn);
   case T42_LOAD_OK:
      InitCodeView(view, view->vm);
      Refresh(view->idl);
      break;
   }
}

static void LoadCommand(void *data)
{
   t_view *view = data;
   const char *fn;

   fn = FileSelect("*.obj;*.lst", "",
                   FS_DIRECT_SELECT_BY_DOUBLE_CLICK|FS_DISABLE_EDIT_DAT|FS_SHOW_MENU|FS_NO_DRAG_DROP|FS_SHOW_DIR_TREE,
                   "Load trac42 object file", "Load");
   if (*fn) {
      LoadFile(fn, view);
   }
}

static void FileMenu(void *data)
{
   t_view *view = data;
   MkMenuItem(0, "Load code", "Ctrl-L", LoadCommand, view);
   MkMenuItem(0, "Exit", "Ctrl-Q", ExitTrac42i, view);
}

static void Help(void *data)
{
   (void)data;
   Req("Help!", "The trac42 virtual machine can be read about in the course "
       "litterature. This actual implementation has a code wrapper at the top of "
       "the vm (addresses 0 and 1). When the vm has finished the execution "
       "the PC will be at address 1. The SP should then be at its original state. "
       "The current PC-line is highlighted as well as breakpoints.| OK ");
}

static void About(void *data)
{
   (void)data;
   Req("About", "Trac42 is a virtual stack machine, similar to Java but a lot simpler. "
       "The trc42 vm has been created by Jukka M�ki-Turja and Christer Norstr�m. "
       "This implementation has been created by Christer Sandberg.| OK ");
}

static void HelpMenu(void *data)
{
   t_view *view = data;
   MkMenuItem(0, "Help", "", Help, view);
   MkMenuItem(0, "About Trac42i", "", About, view);
}

static int CodeLine(void *data, char *s)
{
   t_line *line = data;
   strcpy(s, line->text);
   if (line->i == VMGetPC(line->view->vm)) {
      CGUI_list_row_b_color = cgui_lblue;
   } else if (line->isbreakpoint) {
      CGUI_list_row_b_color = cgui_red;
   }
   return 0;
}

static int StackLine(void *data, char *s, int colnr)
{
   t_stackline *stackline = data;

   switch (colnr) {
   case 0:
      strcpy(s, stackline->regs);
      break;
   case 1:
      sprintf(s, "%d", stackline->address);
      break;
   case 2:
      sprintf(s, "%d", stackline->value);
      break;
   case 3:
      strcpy(s, stackline->label);
      break;
   }
   return 0;
}

static void ClickCodeLine(int id, void *data)
{
   t_line *line = data;

   if (line->isbreakpoint) {
      ClearBreakPointCommand(id, line);
   } else {
      SetBreakPointCommand(line->view, line->i);
   }
}

static void *CreateCodeLine(void *list, int i)
{
   t_view *view=list;
   return view->line + i;
}

static void *CreateStackLine(void *list, int i)
{
   t_view *view=list;
   return view->stacklines + i;
}

int main (int n, char *argv[])
{
   t_view *view;
   int widths[]={40, 80, 80, 80};
   char *labeldata[] = {"Reg","Addr","Val","Label"};
   char **labels = labeldata;

   view = CreateCodeView();
   InitCguiWindowedMode();
   set_window_title("Trac42 interpreter");
   view->vm = VMCreate(GetInput, DisplayOutput, view);
   if (n > 1) {
      LoadFile(argv[1], view);
   }
   view->idw = MkDialogue(FILLSCREEN, NULL, 0);
   MakeMenuBar();
   MkMenuBarItem("File", FileMenu, view);
   MkMenuBarItem("Commands", CommandsMenu, view);
   MkMenuBarItem("Help", HelpMenu, view);
   EndMenuBar();
   view->idl = AddList(DOWNLEFT, view, &view->n, 300, LEFT_MOUSE|RIGHT_MOUSE, CodeLine, ClickCodeLine, 50);
   SetIndexedList(view->idl, CreateCodeLine);
   SetDeleteHandler(view->idl, ClearBreakPointCommand);
   SetInsertHandler(view->idl, SetBreakPointCommand);
   view->idstack = AddList(RIGHT, view, &view->nstacklines, 200, 0, NULL, NULL, 20);
   SetIndexedList(view->idstack, CreateStackLine);
   SetListColumns(view->idstack, StackLine, widths, 4, LIST_COLUMNS_DELIMITER| LIST_COLUMNS_ADJUSTABLE, labels, NULL, NULL);

   DisplayWin();
   SetHotKey(0, StepCommand, view, KEY_F7, 0);
   SetHotKey(0, RunCommand, view, KEY_F9, 0);
   SetHotKey(0, ResetCommand, view, KEY_F2, 0);
   SetHotKey(0, LoadCommand, view, 0, KEY_L);
   SetHotKey(0, ExitTrac42i, view, 0, KEY_Q);
   ProcessEvents();
   return 0;
}
END_OF_MAIN()
