/* $Id: vm.C 98 2008-01-08 18:12:03Z csg $ */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <allegro.h>
#include <cgui.h>
#include <cgui/mem.h>
#include "trac42.h"
#include "vm.h"
#include "instruction.h"
#include "emit.h"

#define STACK_SIZE      1000

#define BUFSIZE 1000

struct t_view;

typedef struct t_block t_block;

struct t_block {
   int size;
   int address;
};

struct t_vm {
   t_inst **code;
   int n;
   int SP;
   int init_SP;
   int PC;
   int FP;
   int OR;
   int *stack;
   int *hai;
   char **labels;
   int start;
   int *bps;
   int nbps;
   t_inst *latest_call;
   struct t_view *view;
   void (*GetInput)(char *, struct t_view*);
   void (*DisplayOutput)(char *, struct t_view*);
};

/* Returns true if the Program counter is on a break point. */
static int ReachedBP(t_vm *vm)
{
  int i;

  for (i=0; i < vm->nbps; i++)
    if (vm->PC == vm->bps[i])
       return 1;
  return 0;
}

static void FillEmptyAddresses(t_vm *vm)
{
   int i;
   for (i=0; i<vm->n; i++) {
      if (vm->code[i] == NULL) {
         InstructionNop(vm, i);
      }
   }
}

/* Reuse memory and initialize the vm to reasonable values. */
static void InitCode(t_vm *vm)
{
   free(vm->code);
   vm->code = NULL;
   vm->n = 0;
   vm->start = 0;
}

extern int VMGetPC(t_vm *vm)
{
   return vm->PC;
}

extern int VMNrOfInstructions(t_vm *vm)
{
   return vm->n;
}

extern int VMFirstInstruction(t_vm *vm, char **src)
{
   return VMNextInstruction(0, vm, src);
}

extern int VMNextInstruction(int handle, t_vm *vm, char **src)
{
   if (handle>=vm->n)
      return END_INSTRUCTIONS;
   InstructionGetSource(vm->code[handle], src);
   return handle + 1;
}

extern int VMNrOfStackElements(t_vm *vm)
{
   return vm->init_SP - vm->SP;
}

extern int VMGetStackBottomElement(t_vm *vm, int *rflags, int *address, int *value, char **label)
{
   return VMGetStackElementAbove(vm->init_SP, vm, rflags, address, value, label);
}

extern int VMGetStackElementAbove(int handle, t_vm *vm, int *rflags, int *address, int *value, char **label)
{
   handle--;
   if (handle < vm->SP)
      return END_STACK_ELEMENTS;
   *rflags = 0;
   if (vm->SP == handle)
      *rflags |= REG_SP;
   if (vm->FP == handle)
      *rflags |= REG_FP;
   *address = handle;
   *value = vm->stack[handle];
   if (vm->labels[handle])
      *label = vm->labels[handle];
   else
      *label = "";
   return handle;
}

extern void VMSetBreakPoint(int address, t_vm *vm)
{
   vm->nbps++;
   vm->bps = ResizeMem(int, vm->bps, vm->nbps);
   vm->bps[vm->nbps-1] = address;
}

extern void VMClearBreakPoint(int address, t_vm *vm)
{
   int i;

   for (i=0; i<vm->nbps; i++) {
      if (vm->bps[i] == address) {
         vm->bps[i] = vm->bps[vm->nbps-1];
         vm->nbps--;
         return;
      }
   }
}

extern void VMClearAllBreakPoints(t_vm *vm)
{
   vm->nbps = 0;
}

extern void VMDestroy(t_vm *vm)
{
   int i;
   for (i=0; i<vm->n; i++) {
      InstructionDestroy(vm->code[i]);
   }
   free(vm->stack);
   free(vm->hai);
   free(vm->code);
   free(vm->labels);
   free(vm->bps);
   free(vm);
}

extern void VMReset(t_vm *vm)
{
   vm->PC = 0;
   vm->FP = STACK_SIZE + STRING_SIZE;
   vm->SP = STACK_SIZE + STRING_SIZE;
   vm->init_SP = vm->SP;
}

extern t_vm *VMCreate(void (*GetInput)(char *, struct t_view *), void (*DisplayOutput)(char *, struct t_view *), struct t_view *view)
{
   t_vm *vm;
   vm = GetMem0(t_vm, 1);
   vm->stack = GetMem0(int, STACK_SIZE + 2*STRING_SIZE);
   vm->hai = GetMem0(int, STACK_SIZE + 2*STRING_SIZE);
   vm->labels = GetMem0(char*, STACK_SIZE + 2*STRING_SIZE);
   vm->view = view;
   vm->DisplayOutput = DisplayOutput;
   vm->GetInput = GetInput;
   return vm;
}

extern int VMExecuteNextInstruction(t_vm *vm)
{
   int status;

   if (vm->PC < vm->n) {
      status = InstructionExecute(vm->code[vm->PC]);
      if (status != EXEC_OK) {
         return status;
      }
   } else {
      return EXEC_ERROR;
   }
   if (vm->PC >= vm->n || vm->PC < 0)
      return EXEC_ERROR_PC;
   if (vm->SP > STACK_SIZE + STRING_SIZE || vm->SP < STRING_SIZE)
      return EXEC_ERROR_SP;
   if (ReachedBP(vm))
      return EXEC_BREAK_POINT;
   return EXEC_OK;
}

extern void VMAddInstruction(t_vm *vm, int instr_address, t_inst *instruction)
{
   int i;
   if (instr_address >= vm->n) {
      vm->code = ResizeMem(t_inst*, vm->code, instr_address + 1);
      for (i=vm->n; i< instr_address + 1; i++) {
         vm->code[i] = NULL;
      }
      vm->n = instr_address + 1;
   }
   vm->code[instr_address] = instruction;
}

extern void VMSetTrac42Address(t_vm *vm, int instr_address)
{
   vm->start = instr_address;
}

extern int VMLoadListFile(const char *fn, t_vm *vm)
{
   int retcode, saved_ret=0;
   char buf[BUFSIZE];
   FILE *f;

   f = fopen (fn, "rt");
   if (f) {
      InitCode(vm);
      do {
         if (fgets(buf, BUFSIZE, f) == NULL) {
            retcode = EOF;
         } else {
            retcode = InstructionParseListLine(buf, vm);
         }
         if (retcode == T42_LOAD_REDUNDANT_CHARS) {
            /* We accept this, it may be unexpected comments, but the code should be OK. */
            saved_ret = T42_LOAD_REDUNDANT_CHARS;
            retcode = T42_LOAD_OK;
         }
      } while (retcode == T42_LOAD_OK);
      fclose(f);
      /* Insert the code wrapper at the beginning of the code. */
      InstructionBsr(vm, 0, vm->start);
      InstructionEnd(vm, 1);

      FillEmptyAddresses(vm);
      VMReset(vm);
      if (retcode == EOF)
         retcode = T42_LOAD_OK;
      if (saved_ret)
         retcode = T42_LOAD_REDUNDANT_CHARS;
      if (vm->start == 0 && retcode == T42_LOAD_OK)
         retcode = T42_LOAD_NO_HANDLE;
   } else {
      retcode =  T42_LOAD_OPEN_FAIL;
   }
   return retcode;
}

extern int VMLoadBinFile(const char *fn, t_vm *vm)
{
   int retcode, n;
   char buf[200];
   FILE *f;
   tInstr instr;

   f = fopen (fn, "rb");
   if (f) {
      InitCode(vm);
      do {
         n = fread(&instr, sizeof (tInstr), 1, f);
         if (n > 0) {
            format_instruction(&instr, buf);
            retcode = InstructionParseListLine(buf, vm);
         }
      } while (n > 0);
      fclose(f);
      /* Insert the code wrapper at the beginning of the code. */
      InstructionBsr(vm, 0, vm->start);
      InstructionEnd(vm, 1);

      FillEmptyAddresses(vm);
      VMReset(vm);
      retcode = T42_LOAD_OK;
      if (vm->start == 0 && retcode == T42_LOAD_OK)
         retcode = T42_LOAD_NO_HANDLE;
   } else {
      retcode =  T42_LOAD_OPEN_FAIL;
   }
   return retcode;
}

/* -------------------------- Execution ------------------- */
extern int VMEnd(t_vm *vm)
{
   if (vm->SP == vm->init_SP)
      return EXEC_CODE_END;
   else
      return EXEC_CODE_END_SP_ERROR;
}

extern int VMLabel(t_vm *vm, t_inst *instruction)
{
   vm->latest_call = instruction;
   vm->PC++;
   return EXEC_OK;
}

extern int VMLink(t_vm *vm)
{
   vm->SP--;
   vm->stack[vm->SP]= vm->FP;
   vm->FP = vm->SP;
   vm->PC++;
   if (vm->latest_call) {
      vm->labels[vm->SP] = InstructionGetStackLabel(vm->latest_call);
      vm->latest_call = NULL;
   } else {
      vm->labels[vm->SP] = NULL;
   }
   return EXEC_OK;
}

extern int VMUnlink(t_vm *vm)
{
   vm->SP = vm->FP;
   vm->FP = vm->stack[vm->SP];
   vm->labels[vm->SP] = NULL;
   vm->SP++;
   vm->PC++;
   return EXEC_OK;
}

extern int VMRts(t_vm *vm)
{
   vm->PC = vm->stack[vm->SP];
   vm->SP++;
   return EXEC_OK;
}

extern int VMDecl(t_vm *vm, int size, char *stack_label)
{
   int i;
   vm->SP -= size;
   for (i=0; i< size; i++) {
      vm->labels[vm->SP+i] = stack_label;
      vm->hai[vm->SP+i] = 0;
   }
   vm->PC++;
   return EXEC_OK;
}

extern int VMLval(t_vm *vm, int *offset, char *stack_label)
{
   vm->SP--;
   if (offset) {
      vm->stack[vm->SP] = vm->FP + *offset;
   } else {
      vm->stack[vm->SP] = vm->FP + vm->OR;
   }
   vm->hai[vm->SP] = 0;
   vm->labels[vm->SP] = stack_label;
   vm->PC++;
   return EXEC_OK;
}

extern int VMAss(t_vm *vm, int size)
{
   int i, stack_address;
   int *dest_address, *hai_dest_address;

   if (vm->hai[vm->SP + size]) {
      dest_address = (int*)vm->stack[vm->SP + size];
      hai_dest_address = dest_address + size;
   } else {
      stack_address = vm->stack[vm->SP + size];
      if (stack_address < vm->SP) {
         return EXEC_ERROR;
      }
      dest_address = vm->stack + stack_address;
      hai_dest_address = vm->hai + stack_address;
   }
   for (i=0; i < size; i++) {
      dest_address[i] = vm->stack[vm->SP + i];
      hai_dest_address[i] = vm->hai[vm->SP + i];
   }
   vm->SP += size + 1;
   vm->PC++;
   return EXEC_OK;
}

extern int VMBrf(t_vm *vm, int target_address)
{
   if (!vm->stack[vm->SP])
      vm->PC = target_address;
   else
      vm->PC++;
   vm->SP++;
   return EXEC_OK;
}

extern int VMBra(t_vm *vm, int target_address)
{
   vm->PC = target_address;
   return EXEC_OK;
}

extern int VMReadString(t_vm *vm)
{
   char str[BUFSIZE], *s;
   int i;

   vm->GetInput(str, vm->view);
   s = str;
   vm->SP -= 100;
   for (i=0; str[i]; i++) {
      vm->stack[vm->SP + i] = str[i];
      vm->labels[vm->SP + i] = NULL;
      vm->hai[vm->SP + i] = 0;
   }
   vm->stack[vm->SP + i] = 0;
   vm->PC++;
   return EXEC_OK;
}

extern int VMReadBool(t_vm *vm)
{
   char str[BUFSIZE], *s;

   vm->GetInput(str, vm->view);
   s = str;
   while (*s && isspace(*s))
      s++;
   vm->SP--;
   vm->labels[vm->SP] = NULL;
   if (stricmp(s, "false")==0)
      vm->stack[vm->SP] = 0;
   else if (stricmp(s, "true")==0)
      vm->stack[vm->SP] = 1;
   else
      sscanf(s, "%d", &vm->stack[vm->SP]);
   vm->hai[vm->SP] = 0;
   vm->PC++;
   return EXEC_OK;
}

extern int VMReadInt(t_vm *vm)
{
   char str[BUFSIZE];

   vm->GetInput(str, vm->view);
   vm->SP--;
   vm->labels[vm->SP] = NULL;
   sscanf(str, "%d", &vm->stack[vm->SP]);
   vm->hai[vm->SP] = 0;
   vm->PC++;
   return EXEC_OK;
}

extern int VMWriteString(t_vm *vm)
{
   int i;
   char s[STRING_SIZE+1];

   for (i=0; i<STRING_SIZE && vm->stack[vm->SP+i]; i++)
      s[i] = vm->stack[vm->SP+i];
   s[i] = 0;
   vm->DisplayOutput(s, vm->view);
   vm->PC++;
   return EXEC_OK;
}

extern int VMWriteBool(t_vm *vm)
{
   char *s;

   if (!vm->stack[vm->SP])
      s = "false";
   else
      s = "true";
   vm->DisplayOutput(s, vm->view);
   vm->PC++;
   return EXEC_OK;
}

extern int VMWriteInt(t_vm *vm)
{
   char s[100];

   sprintf(s, "%d",vm->stack[vm->SP]);
   vm->DisplayOutput(s, vm->view);
   vm->PC++;
   return EXEC_OK;
}

extern int VMBsr(t_vm *vm, int target_address)
{
   vm->SP--;
   vm->labels[vm->SP] = NULL;
   vm->stack[vm->SP] = vm->PC + 1;
   vm->hai[vm->SP] = 0;
   vm->PC = target_address;
   return EXEC_OK;
}

extern int VMPop(t_vm *vm, int size)
{
   vm->SP += size;
   vm->PC++;
   return EXEC_OK;
}

extern int VMNot(t_vm *vm)
{
   vm->stack[vm->SP] = !vm->stack[vm->SP];
   vm->labels[vm->SP] = NULL;
   vm->hai[vm->SP] = 0;
   vm->PC++;
   return EXEC_OK;
}

extern int VMNeg(t_vm *vm)
{
   vm->stack[vm->SP] = -vm->stack[vm->SP];
   vm->labels[vm->SP] = NULL;
   vm->hai[vm->SP] = 0;
   vm->PC++;
   return EXEC_OK;
}

extern int VMOr(t_vm *vm)
{
   vm->stack[vm->SP + 1] = vm->stack[vm->SP + 1] || vm->stack[vm->SP];
   vm->SP++;
   vm->labels[vm->SP] = NULL;
   vm->hai[vm->SP] = 0;
   vm->PC++;
   return EXEC_OK;
}

extern int VMAnd(t_vm *vm)
{
   vm->stack[vm->SP + 1] = vm->stack[vm->SP + 1] && vm->stack[vm->SP];
   vm->SP++;
   vm->labels[vm->SP] = NULL;
   vm->hai[vm->SP] = 0;
   vm->PC++;
   return EXEC_OK;
}

extern int VMEqString(t_vm *vm)
{
   int i = 0;
   char left[STRING_SIZE], right[STRING_SIZE];

   while (i < (STRING_SIZE - 1)) {
      left[i] = vm->stack[vm->SP + STRING_SIZE + i];
      right[i] = vm->stack[vm->SP + i];
      i++;
   }
   vm->SP += 2 * STRING_SIZE - 1;
   vm->stack[vm->SP] = !strncmp(left, right, (STRING_SIZE - 1));
   vm->labels[vm->SP] = NULL;
   vm->hai[vm->SP] = 0;
   vm->PC++;
   return EXEC_OK;
}

extern int VMEqInt(t_vm *vm)
{
   vm->stack[vm->SP + 1] = vm->stack[vm->SP + 1] == vm->stack[vm->SP];
   vm->SP++;
   vm->labels[vm->SP] = NULL;
   vm->hai[vm->SP] = 0;
   vm->PC++;
   return EXEC_OK;
}

extern int VMLtString(t_vm *vm)
{
   int i = 0;
   char left[STRING_SIZE], right[STRING_SIZE];

   while (i < (STRING_SIZE - 1)) {
      left[i] = vm->stack[vm->SP + STRING_SIZE + i];
      right[i] = vm->stack[vm->SP + i];
      i++;
   }
   vm->SP += 2 * STRING_SIZE - 1;
   vm->stack[vm->SP] = strncmp(left, right, (STRING_SIZE - 1)) < 0;
   vm->labels[vm->SP] = NULL;
   vm->hai[vm->SP] = 0;
   vm->PC++;
   return EXEC_OK;
}

extern int VMLtInt(t_vm *vm)
{
   vm->stack[vm->SP + 1] = vm->stack[vm->SP + 1] < vm->stack[vm->SP];
   vm->SP++;
   vm->labels[vm->SP] = NULL;
   vm->hai[vm->SP] = 0;
   vm->PC++;
   return EXEC_OK;
}

extern int VMLeString(t_vm *vm)
{
   int i = 0;
   char left[STRING_SIZE], right[STRING_SIZE];

   while (i < (STRING_SIZE - 1)) {
      left[i] = vm->stack[vm->SP + STRING_SIZE + i];
      right[i] = vm->stack[vm->SP + i];
      i++;
   }
   vm->SP += 2 * STRING_SIZE - 1;
   vm->stack[vm->SP] = strncmp(left, right, (STRING_SIZE - 1)) <= 0;
   vm->labels[vm->SP] = NULL;
   vm->hai[vm->SP] = 0;
   vm->PC++;
   return EXEC_OK;
}

extern int VMLeInt(t_vm *vm)
{
   vm->stack[vm->SP + 1] = vm->stack[vm->SP + 1] <= vm->stack[vm->SP];
   vm->SP++;
   vm->labels[vm->SP] = NULL;
   vm->hai[vm->SP] = 0;
   vm->PC++;
   return EXEC_OK;
}

extern int VMSub(t_vm *vm)
{
   vm->stack[vm->SP + 1] -= vm->stack[vm->SP];
   vm->hai[vm->SP + 1] |= vm->hai[vm->SP];
   vm->SP++;
   vm->labels[vm->SP] = NULL;
   vm->PC++;
   return EXEC_OK;
}

extern int VMAdd(t_vm *vm)
{
   vm->stack[vm->SP + 1] += vm->stack[vm->SP];
   vm->hai[vm->SP + 1] |= vm->hai[vm->SP];
   vm->SP++;
   vm->labels[vm->SP] = NULL;
   vm->PC++;
   return EXEC_OK;
}

extern int VMMult(t_vm *vm)
{
   vm->stack[vm->SP + 1] *= vm->stack[vm->SP];
   vm->SP++;
   vm->labels[vm->SP] = NULL;
   vm->hai[vm->SP] = 0;
   vm->PC++;
   return EXEC_OK;
}

extern int VMDiv(t_vm *vm)
{
   if (vm->stack[vm->SP] == 0)
      return EXEC_ERROR_DIV_ZERO;
   vm->stack[vm->SP + 1] /= vm->stack[vm->SP];
   vm->SP++;
   vm->labels[vm->SP] = NULL;
   vm->hai[vm->SP] = 0;
   vm->PC++;
   return EXEC_OK;
}

extern int VMPushString(t_vm *vm, char *string)
{
   int i = 0;

   vm->SP -= STRING_SIZE;
   while (string[i] && i < (STRING_SIZE - 1)) {
      vm->stack[vm->SP + i] = string[i];
      vm->labels[vm->SP + i] = NULL;
      vm->hai[vm->SP + i] = 0;
      i++;
   }
   vm->stack[vm->SP + i] = 0;
   vm->PC++;
   return EXEC_OK;
}

extern int VMPushInt(t_vm *vm, int value)
{
   vm->SP--;
   vm->stack[vm->SP] = value;
   vm->labels[vm->SP] = NULL;
   vm->hai[vm->SP] = 0;
   vm->PC++;
   return EXEC_OK;
}

extern int VMRval(t_vm *vm, int *offset, int size)
{
   int i, off;

   if (offset) {
      off = *offset;
   } else {
      off = vm->OR;
   }
   vm->SP -= size;
   for (i = 0; i < size; i++) {
      vm->stack[vm->SP + i] = vm->stack[vm->FP + off + i];
      vm->hai[vm->SP + i] = vm->hai[vm->FP + off + i];
      vm->labels[vm->SP + i] = NULL;
   }
//    vm->labels[vm->SP] = InstructionGetStackLabel(instruction);
   vm->PC++;
   return EXEC_OK;
}

extern int VMDeref(t_vm *vm, int size, char *stack_label)
{
   int i, stack_address;
   int *source_address, *hai_source_address;

   if (vm->hai[vm->SP]) {
      source_address = (int*)vm->stack[vm->SP];
      hai_source_address = source_address + size;
   } else {
      stack_address = vm->stack[vm->SP];
      if (stack_address < STRING_SIZE) {
         return EXEC_ERROR;
      }
      source_address = vm->stack + stack_address;
      hai_source_address = vm->hai + stack_address;
   }
   vm->SP++;
   vm->SP -= size;
   for (i = 0; i < size;  i++) {
      vm->stack[vm->SP + i] = source_address[i];
      vm->hai[vm->SP + i] = hai_source_address[i];
   }
   vm->labels[vm->SP] = stack_label;
   vm->PC++;
   return EXEC_OK;
}

extern int VMSetOffset(t_vm *vm)
{
   vm->OR = vm->stack[vm->SP];
   vm->SP++;
   vm->PC++;
   return EXEC_OK;
}

extern int VMAlloc(t_vm *vm, int size)
{
   int *p;

   p = calloc(size*2, sizeof(int));
   if (p == NULL) {
      return EXEC_ERROR;
   }
   vm->SP--;
   vm->stack[vm->SP] = (int)p;
   vm->hai[vm->SP] = 1;
   vm->PC++;
   return EXEC_OK;
}

/* -------------------------- END Execution ------------------- */
