/* $Id: vm.h 98 2008-01-08 18:12:03Z csg $ */
#ifndef _VM_H_
#define _VM_H_

struct t_view;
struct t_inst;
typedef struct t_vm t_vm;

/* Returns the total number of loaded instructions. */
extern int VMNrOfInstructions(t_vm *vm);

/* Returns the total number of elements that is currently on the stack. */
extern int VMNrOfStackElements(t_vm *vm);

/* Executes the instruction pointed to by the current Program counter. Returns either of the codes below to indicate
   the success of the execution. */
extern int VMExecuteNextInstruction(t_vm *vm);
#define EXEC_ERROR_INSTR_CODE 1
#define EXEC_ERROR_PC         2
#define EXEC_ERROR_SP         3
#define EXEC_ERROR_DIV_ZERO   4
#define EXEC_BREAK_POINT      5
#define EXEC_CODE_END_SP_ERROR 6
#define EXEC_CODE_END         7
#define EXEC_OK               8
#define EXEC_ERROR            9

/* Sets a break point at `address'. When executing in full speed the execution will stop when reaching that address. */
extern void VMSetBreakPoint(int address, t_vm *vm);

/* Clears a break point set at `address' if there is one, else do nothing. */
extern void VMClearBreakPoint(int address, t_vm *vm);

/* Clears all break points. */
extern void VMClearAllBreakPoints(t_vm *vm);

/* Frees all memeory associated with a virtual machine as well as the memory pointed to by vm. After the return
   from a call to `DestroyVirtualMachine' `vm' is no longer a valid pointer. */
extern void VMDestroy(t_vm *vm);

/* Reset the state of the program. Used stack space will not be updated. */
extern void VMReset(t_vm *vm);

/* Returns a pointer to a new virtual machine object. It will store the pointers GetInput and DisplayOutput. These will be called each time
   an input or output (currently read and write) instruction is executed.
   While GetInput is called the execution is suspended. After the call the string at the location pointed to by `s' will be used to
   store a new value in the virtual macine.
   While DisplayOutput is called the execution is suspended. The string pointed to by `s' can be used by DisplayOutput to display the
   output from the virtual machine.
*/
extern t_vm *VMCreate(void (*GetInput)(char *s, struct t_view *view), void (*DisplayOutput)(char *s, struct t_view *view), struct t_view *view);

/* Loads code into the virtual machine from a list file. Returns either of the codes below to indicate the success of the
   loading. */
extern int VMLoadListFile(const char *fn, t_vm *vm);
extern int VMLoadBinFile(const char *fn, t_vm *vm);
#define T42_LOAD_OPEN_FAIL       1
#define T42_LOAD_NO_HANDLE       2
#define T42_LOAD_FORMAT_ERROR    3
#define T42_LOAD_REDUNDANT_CHARS 4
#define T42_LOAD_OK              5

/* Returns the address of the program counter. */
extern int VMGetPC(t_vm *vm);

/* Iterator functions over the stack. */

/* Returns an iterator to the bottom element of the stack or END_STACK_ELEMENTS if there was no element.
   If there was an element the returned handle can be used for input to GetStackElementAbove and
   the location pointed to by `rflags' will get the flag REG_SP set if this is the stack element pointed to
   by the stack pointer (i.e. the last) and with the flags REG_FP if this is the stack element pointed to
   by the frame pointer and
   the location pointed to by `address' will be updated with the memory address of this stack element and
   the location pointed to by `value' will be updated with the content of this stack element and
   the location pointed to by `label' will be updated with a pointer that contains a (possibly empty) stack label.
   `vm' is a pointer to the virtual machine. */
extern int VMGetStackBottomElement(t_vm *vm, int *rflags, int *address, int *value, char **label);
/* Returns an iterator to the next stack element above `hnd' or END_INSTRUCTIONS if there was no more instructions.
   If there was an element the returned handle can be used for input to another call to GetStackBottomElement and
   the locations pointed to by the parameters will be updated with values as for GetStackBottomElement. */
extern int VMGetStackElementAbove(int handle, t_vm *vm, int *rflags, int *address, int *value, char **label);
#define END_STACK_ELEMENTS    0
#define REG_SP    (1<<0)
#define REG_FP    (1<<1)

/* Iterator functions over the instuctions. */

/* Returns an iterator to the first instruction or END_STACK_ELEMENTS if there was no more elements.
   If there was an instruction the returned handle can be used for input to NextInstruction and
   the location pointed to by `asm_code' will be updated with a pointer that contains a textual representation of the instruction.
   `vm' is a pointer to the virtual machine. */
extern int VMFirstInstruction(t_vm *vm, char **asm_code);
/* Returns an iterator to the next instruction after `hnd' or END_INSTRUCTIONS if there was no more instructions.
   If there was an instruction the returned handle can be used for input to another call to NextInstruction and
   the location pointed to by `asm_code' will be updated with a pointer that contains a textual representation of the instruction.
   `vm' is a pointer to the virtual machine. */
extern int VMNextInstruction(int handle, t_vm *vm, char **asm_code);

/* Adds `instruction' to the virtual machine's program memory at instr_address. */
extern void VMAddInstruction(t_vm *vm, int instr_address, struct t_inst *instruction);

/* Set the adress to the trac42 function. */
extern void VMSetTrac42Address(t_vm *vm, int instr_address);

/* The various instructions. */
extern int VMEnd(t_vm *vm);
extern int VMLabel(t_vm *vm, struct t_inst *instruction);
extern int VMLink(t_vm *vm);
extern int VMUnlink(t_vm *vm);
extern int VMRts(t_vm *vm);
extern int VMDecl(t_vm *vm, int size, char *stack_label);
/* If offset is NULL the current value of the offset register, OR, will be used as offset. */
extern int VMLval(t_vm *vm, int *offset, char *stack_label);
extern int VMAss(t_vm *vm, int size);
extern int VMBrf(t_vm *vm, int target_address);
extern int VMBra(t_vm *vm, int target_address);
extern int VMReadString(t_vm *vm);
extern int VMReadBool(t_vm *vm);
extern int VMReadInt(t_vm *vm);
extern int VMWriteString(t_vm *vm);
extern int VMWriteBool(t_vm *vm);
extern int VMWriteInt(t_vm *vm);
extern int VMBsr(t_vm *vm, int target_address);
extern int VMPop(t_vm *vm, int size);
extern int VMNot(t_vm *vm);
extern int VMNeg(t_vm *vm);
extern int VMOr(t_vm *vm);
extern int VMAnd(t_vm *vm);
extern int VMEqString(t_vm *vm);
extern int VMEqInt(t_vm *vm);
extern int VMLtString(t_vm *vm);
extern int VMLtInt(t_vm *vm);
extern int VMLeString(t_vm *vm);
extern int VMLeInt(t_vm *vm);
extern int VMSub(t_vm *vm);
extern int VMAdd(t_vm *vm);
extern int VMMult(t_vm *vm);
extern int VMDiv(t_vm *vm);
extern int VMPushString(t_vm *vm, char *string);
extern int VMPushInt(t_vm *vm, int value);
extern int VMRval(t_vm *vm, int *offset, int size);
extern int VMDeref(t_vm *vm, int size, char *stack_label);
extern int VMAlloc(t_vm *vm, int size);
/* Pops the offset from the stack and set the offset register, OR, to that value. */
extern int VMSetOffset(t_vm *vm);

#define END_INSTRUCTIONS     -1

#endif
